import Component from '@ember/component';
import EmberObject from '@ember/object';
export declare enum SelectorStrategies {
    PARENT = "parent",
    PARENT_VIEW = "parentView",
    SELF = "element"
}
export declare enum TriggerEvents {
    CLICK = "click",
    HOVER = "hover",
    FOCUS = "focus"
}
declare class UiContextContainerInState extends EmberObject {
    [TriggerEvents.HOVER]: boolean;
    [TriggerEvents.FOCUS]: boolean;
    [TriggerEvents.CLICK]: boolean;
    get show(): boolean;
}
/**
 * A base class for associating "reference" and "overlay" DOM elements with each other, as
 * would be needed for tooltips, popovers, floating menus, etc. Used in close coordination with
 * UiContextualElement, and UiPopper.
 */
export default class UiContextualContainer extends Component {
    static readonly positionalParams: string[];
    /**
     * Placement of the positioned element, relative to its reference element.
     * One of ['top', 'right', 'bottom', 'left'].
     */
    placement: 'top' | 'right' | 'bottom' | 'left';
    /**
     * If true, the positioned element will change its position automatically based
     * on available space.
     */
    autoPlacement: boolean;
    /**
     * Whether the component is accepting any interactions.
     */
    enabled: boolean;
    /**
     * Whether the positioned element is in view.
     *
     * @hidden
     */
    visible: boolean;
    /**
     * When triggerEvents = CLICK, this controls whether interacting with elements
     * outside the positioned or triggering elements will cause the positioned
     * element to be hidden.
     */
    closeOnOutsideClick: boolean;
    /**
     * The amount of time, in milliseconds, between the trigger event occurring
     * and the positioned element being shown or hidden. This value is applied
     * equally to both show and hide events. See `delayShow` and `delayHide` to
     * control each value independently.
     */
    delay: number;
    /**
     * The amount of time, in milliseconds, between the trigger event occurring
     * and the positioned element being shown. If not explicitly set, this
     * will be equal to the value of `delay`.
     *
     * @hidden
     */
    delayShow: number;
    /**
     * The amount of time, in milliseconds, between the trigger event occurring
     * and the positioned element being hidden. If not explicitly set, this
     * will be equal to the value of `delay`.
     *
     * @hidden
     */
    delayHide: number;
    /**
     * A query selector for the container describing the area that the positioned
     * element will be checked for overflow relative to.
     *
     * @see https://popper.js.org/docs/v2/modifiers/prevent-overflow/#boundary
     *
     * @hidden
     */
    viewportSelector: string;
    /**
     * Virtual "padding" for the PopperJS preventOverflow modifier.
     *
     * @see https://popper.js.org/docs/v2/modifiers/prevent-overflow/#padding
     *
     * @hidden
     */
    viewportPadding: number | {
        top?: number;
        left?: number;
        right?: number;
        bottom?: number;
    };
    /**
     * An offset, in pixels, for the PopperJS offset modifier.
     *
     * @see https://popper.js.org/docs/v2/modifiers/offset/#distance
     */
    distance: number;
    /**
     * If true, the element will be displayed and positioned within its parent element.
     * When false, the element will be moved to a higher position in the DOM. Use this
     * to mitigate z-index and other layout issues.
     */
    renderInPlace: boolean;
    /**
     * Describes the type of action(s) that will cause the positioned element to be
     * shown/hidden. Possible values are "hover", "focus", and "click". These can be
     * provided as an array, or a single string with a space between each option.
     *
     * @hidden
     */
    triggerEvents: TriggerEvents | TriggerEvents[];
    /**
     * A query selector or keyword for the DOM element that will be acting as the
     * "trigger" for showing/hiding the positioned element. Available keywords:
     *
     * - "self":       The root element created by this component if a tagName is given.
     * - "parent":     The DOM element that this component is created inside.
     * - "parentView": The Ember Component instance that this component is created inside.
     *
     * @hidden
     */
    triggerSelector: SelectorStrategies;
    /**
     * A query selector or keyword for the DOM element that will be associated with the
     * content of the positioned element for accessibility purposes. E.g. a tooltip may
     * be triggered by a button, but its content describe a form field label. In this
     * case, the button is the trigger element, and the form field label is the ARIA
     * element. Available keywords:
     *
     * - "self":       The root element created by this component if a tagName is given.
     * - "parent":     The DOM element that this component is created inside.
     * - "parentView": The Ember Component instance that this component is created inside.
     * - "trigger":    The DOM "trigger" element.
     *
     * Set as null/undefined if no such support is desired.
     *
     * @hidden
     */
    ariaSelector?: string;
    /**
     * The ARIA attribute that will be used to associate the positioned element with the
     * `ariaSelector` element.
     */
    ariaAttachAs: string;
    /**
     * Whether to use a fade transition on the positioned element as it is being
     * shown/hidden.
     */
    fade: boolean;
    /**
     * The approximate number, in milliseconds, that the fade transition runs. This is
     * a backup just in case "transitionEnd" is not supported and shouldn't ever need
     * to be fiddled with.
     *
     * @hidden
     */
    animationDuration: number;
    /**
     * A method that will execute when the positioned element first begins to
     * visibly appear onscreen. Returning a boolean false from this method will
     * prevent the positioned element from being shown.
     */
    onShow?: (instance: UiContextualContainer) => void | boolean;
    /**
     * A method that will execute when the positioned element is visible onscreen.
     * This will be after all transitions have finished.
     */
    onShown?: (instance: UiContextualContainer) => void;
    /**
     * A method that will execute when the positioned element first begins to
     * disappear offscreen. Returning a boolean false from this method will
     * prevent the positioned element from being removed.
     */
    onHide?: (instance: UiContextualContainer) => void | boolean;
    /**
     * A method this will execute when the positioned element has disappeared
     * offscreen. This will be after all transitions have finished.
     */
    onHidden?: (instance: UiContextualContainer) => void;
    /**
     * A unique ID for the positioned element.
     */
    overlayId?: string;
    /**
     * An optional value for the CSS max-width property of the positioned element.
     */
    maxWidth?: string | number;
    /**
     * The name of the component that will be rendered as the positioned element.
     * It should always be something that extends UiContextElement.
     */
    overlayComponent: string;
    /**
     * Text that will put in the body of the rendered `overlayComponent` if not yielding
     * block content. This is a pretty simple convenience that allows UiContextContainer
     * to be quickly used inline.
     */
    textContent?: string;
    /**
     * Text that will put in the header of the rendered `overlayComponent`. This is a pretty
     * simple convenience that allows UiContextContainer to be quickly used inline.
     */
    title?: string;
    /**
     * The _isHidden and _showOverlay properties work together to display the positioned
     * element. When being transitioned in, _isHidden will toggle `true` first, indicating
     * that the element should be rendered. Once rendered, _showOverlay is toggled `true`
     * as well, which causes the CSS transition in to begin. Hiding is simply the reverse.
     *
     * @private
     */
    _isHidden: boolean;
    /**
     * @private
     */
    _showOverlay: boolean;
    /**
     * Boolean indicating whether a show delay has been set.
     */
    protected readonly hasDelayShow: boolean;
    /**
     * Boolean indicating whether a hide delay has been set.
     */
    protected readonly hasDelayHide: boolean;
    /**
     * The run loop timer ID for the delayShow and delayHide pauses between when
     * an interaction occurs and the positioned element beings being transitioned.
     */
    private timer?;
    /**
     * A state management flag used to help marshall show/hide the positioned
     * element for mouseover and focus events.
     */
    private hoverState?;
    /**
     * The EventListener that is set up to handle mouseover and focus events.
     */
    protected handleEnter: EventListener;
    /**
     * The EventListener that is set up to handle mouseleave and blur events.
     */
    protected handleLeave: EventListener;
    /**
     * The EventListener that is set up to handle click events.
     */
    protected handleToggle: EventListener;
    /**
     * The EventListener that is set up on the document when triggerEvents
     * contains "click". This listener is responsible for closing the overlay
     * when an outside click occurs.
     */
    protected handleOutsideClick?: EventListener;
    /**
     * A Text node inserted into the template to find the parent element of
     * this component when it itself is tagless.
     */
    private parentFinder;
    /**
     * The unique ID of the positioned element which will be generated
     * if not provided via `overlayId`.
     */
    protected get overlayElementId(): string;
    /**
     * The container to render the positioned element into if `renderInPlace`
     * is false. This container is provided by the addon itself, injected into
     * the application's index.html at build time.
     */
    protected get overlayDestinationElement(): HTMLElement;
    /**
     * The trigger element is what gets interacted with to activate
     * the overlay. This is purposefully memoized with no dependent
     * keys - once set the trigger should not change for a single
     * component instance.
     */
    protected get triggerElement(): HTMLElement | null;
    /**
     * A more thorough `renderInPlace` used internally by the component that
     * also takes into account whether there is a destination container
     * to worm-hole the positioned element out to.
     */
    protected get actuallyRenderInPlace(): boolean;
    /**
     * The HTMLElement result of querying for the value of `viewportSelector`.
     */
    protected get viewportElement(): Element | null;
    /**
     * Yet another state management doodad.
     */
    protected get inState(): UiContextContainerInState;
    /**
     * An array of provided `triggerEvents`.
     */
    protected get triggerEventsArray(): TriggerEvents[];
    /**
     * @protected
     * */
    init(): void;
    /**
     * @protected
     */
    didInsertElement(): void;
    /**
     * @protected
     * */
    willDestroyElement(): void;
    /**
     * The overlay element is the "attached" thing - e.g. the DOM
     * of the tooltip, flyout, menu, etc.
     */
    protected getOverlayElement(): HTMLElement | null;
    /**
     * The ARIA element is some additional DOM node to be specially
     * annotated for increased accessibility.
     */
    protected getAriaElement(): HTMLElement | null;
    /**
     * Used for the event listeners attached to the trigger element to short-circuit events that
     * happen from within the overlay. Not doing this could cause weird things to happen if the
     * overlay is being rendered in-place as a child of the trigger.
     */
    protected handleTriggerEvent(handler: (event: Event) => void, e: Event): any;
    /**
     * Attach event listeners to the provided element. This occurs for the
     * trigger element during component setup, and for the positional element
     * when it is shown.
     */
    protected attachTriggerEventListeners(target: HTMLElement | null, triggers: TriggerEvents[]): void;
    /**
     * Remove event listeners from the provided element. This occurs for the
     * trigger element during component teardown, and for the positional element
     * when it is hidden.
     */
    protected removeTriggerEventListeners(target: HTMLElement | null, triggers: TriggerEvents[]): void;
    /**
     * Manages the state change of the overlay from hidden to visible.
     */
    protected show(): void;
    /**
     * Queued into the run loop by show(), this does the actual transition work.
     */
    protected _show(skipAnimation?: boolean): void;
    /**
     * Manages the state change of the overlay from visible to hidden.
     */
    protected hide(skipAnimation?: boolean): void;
    /**
     * Internal state management used by the hover + focus trigger event scheme
     * to make the overlay visible if it is not already.
     */
    protected enter(event?: Event): void;
    /**
     * Internal state management used by the hover + focus trigger event scheme
     * to make the overlay hidden if it is not already.
     */
    protected leave(event?: Event): void;
    /**
     * Internal state management used by the click trigger event scheme to
     * toggle the visibility of the overlay.
     */
    protected toggle(event?: Event): void;
    /**
     * Yielded action to programmatically close from within the tooltip/flyout/whatever.
     */
    protected close(): void;
    /**
     * Adds the ARIA property to the target element.
     */
    protected attachAriaDescriptor(): void;
    /**
     * Cleans the ARIA property of the target element.
     */
    protected removeAriaDescriptor(): void;
    /**
     * Convenience method to check whether we should continue to do...
     * pretty much anything with this component instance.
     */
    protected checkDestruction(): boolean;
    /**
     * Observer change handler of the "visible" property.
     */
    protected onVisiblePropertyChange(): void;
    /**
     * Observer change handler of the "enabled" property.
     */
    protected onEnabledPropertyChange(): void;
    /**
     * Looks up an HTMLElement in a variety of ways. Used by this component, for example,
     * to find the trigger element, the ARIA target element, etc.
     */
    protected findElementWithSelectorStrategy(strategy: SelectorStrategies | string): HTMLElement | null;
}
export {};
