import type { Modifier } from '@popperjs/core';
import Component from '@ember/component';
/**
 * The UiContextElement component serves as the base class for tooltips, popovers, menus,
 * and more. Anything that needs to be dynamically positioned relative to some other "anchor"
 * element begins here.
 *
 * @class UiContextualElement
 */
export default class UiContextualElement extends Component {
    /**
     * The role attribute of the element.
     */
    ariaRole: string;
    /**
     * Placement of the element, relative to its reference element.
     * One of ['top', 'right', 'bottom', 'left'].
     */
    placement: string;
    /**
     * Whether the element is visible and positioned.
     */
    isHidden: boolean;
    /**
     * Optional text that will be shown in a separate block above the primary content. Some
     * controls, such as popovers, have an optional title block.
     */
    titleText?: string;
    fade: boolean;
    showContent: boolean;
    /**
     * If true, the element will be displayed and positioned within its parent element. When
     * false, the element will be moved to a higher position in the DOM. Use this to mitigate
     * z-index and other layout issues.
     */
    renderInPlace: boolean;
    /**
     * If true, the element will change its position automatically based on available space.
     */
    autoPlacement: boolean;
    /**
     * Virtual "padding" for the PopperJS preventOverflow modifier.
     * https://popper.js.org/docs/v2/modifiers/prevent-overflow/#padding
     */
    viewportPadding: number | {
        top?: number;
        right?: number;
        bottom?: number;
        left?: number;
    };
    /**
     * An offset, in pixels, for the PopperJS offset modifier.
     * https://popper.js.org/docs/v2/modifiers/offset/#distance
     */
    distance: number;
    /**
     * The "reference" that this element will be positioned relative to.
     */
    popperTarget?: HTMLElement;
    /**
     * Describes the area that the element will be checked for overflow relative to.
     * https://popper.js.org/docs/v2/modifiers/prevent-overflow/#boundary
     */
    viewportElement?: HTMLElement | 'clippingParents';
    /**
     * If `renderInPlace` is false, then this will need to be supplied as the
     * container that this element will be rendered into.
     */
    destinationElement?: HTMLElement;
    testId?: string;
    /**
     * A method provided by the parent container to dismiss the positioned element.
     */
    readonly close: () => void;
    /**
     * If true, a floating close button will be rendered at the top of the positioned
     * element. This can be handy when using the "click" trigger.
     */
    showCloseButton: boolean;
    /**
     * CSS class names for the element.
     */
    get popperClassNames(): undefined | string | string[];
    /**
     * A unique id for the element. One will be generated if not provided.
     */
    id?: string;
    /**
     * An optional value for the CSS max-width property of the inner block of this
     * element.
     */
    maxWidth?: string | number;
    /**
     * If true, a directional caret will be shown, pointing to the target element.
     */
    showArrow: boolean;
    /**
     * CSS class name for the arrow caret.
     */
    arrowClassName: string;
    /**
     * CSS class name for the inner block content of this element.
     */
    innerClassName: string;
    /**
     * CSS class name for the title text content block.
     */
    titleTextClassName?: string;
    /**
     * The real placement of the element. If `autoPlacement` is enabled then
     * this could differ from `placement`.
     */
    readonly actualPlacement: string;
    /**
     * The `placement` property allows for either PopperJS or Bootstrap
     * standard layout position strings to be provided. This readonly
     * value will always be the PopperJS translated version.
     */
    get popperPlacement(): string;
    /**
     * Concatenated CSS class names for the element.
     */
    get popperClasses(): string;
    /**
     * An array of PopperJS modifier descriptions created from the configuration
     * properties of this component.
     */
    get popperModifiers(): Modifier<unknown, Record<string, unknown>>[];
    get innerElementStyles(): import("@ember/template/-private/handlebars").SafeString | undefined;
    protected getPopperModifiers(): Modifier<unknown, Record<string, unknown>>[];
    /**
     * PopperJS uses slightly different terminology than Bootstrap.
     */
    protected updatePlacement(data?: {
        state?: {
            placement: string;
        };
    }): void;
}
