import type { AlertLevel, AlertGroupDefinition } from '@nsf-open/ember-ui-foundation/constants';
import Component from '@ember/component';
import { AlertLevelKeys } from '@nsf-open/ember-ui-foundation/constants';
/**
 * A block notification banner, also commonly referred to as a "page-level alert". There are four
 * variants of banner:
 *
 * * success
 * * warning
 * * danger
 * * secondary (informational)
 *
 * Single alert messages, and arrays of strings are both supported by the component's inline form.
 *
 * ```handlebars
 *
 * {{ui-alert "success" "Hurray, it worked! An array of strings would have worked here too!"}}
 * ```
 *
 * Block content is also supported. To keep consistency across the platform, you will need to place
 * the alert's "title" yourself.
 *
 * ```handlebars
 * <UiAlert @variant="success" as |Alert|>
 *   <p><Alert.title /> Great job team!</p>
 * </UiAlert>
 *
 * {{#ui-alert "success" as |alert|}}
 *   {{alert.title plural=true}}
 *
 *   <ul>
 *     <li>Task A worked!</li>
 *     <li>Task B rocked!</li>
 *   </ul>
 * {{/ui-alert}}
 * ```
 *
 * @yield ({ title: UiAlertTitle }) Alert
 * */
export default class UiAlert extends Component {
    static readonly positionalParams: string[];
    readonly ariaRole = "alert";
    /**
     * The style of alert banner to create. For historical reasons and maximum cross-compatibility
     * there are a number for aliases for each variant.
     *
     * - success, successes
     * - warning, warnings
     * - danger, error, errors, alert
     * - info, informational, informationals, secondary
     * - default, muted
     */
    variant?: AlertLevelKeys;
    /**
     * The text content of the alert banner. If provided and array of multiple strings
     * they will be formatted as an unordered list in the order supplied. This may also be
     * set as the second positional parameter of the component.
     */
    content?: string | string[];
    /**
     * The value of the element's `data-test-ident` attribute, if required. By default,
     * it will be a string that follows the pattern "context-message-${variant}".
     */
    testId?: string;
    /**
     * An object whose keys are AlertLevel strings and corresponding value is an object
     * which can contain `icon`, `singular` and `plural` strings and will be used to
     * create the alert heading. Levels that are not defined here will fall back to the
     * defaults.
     */
    alertGroups?: Record<AlertLevel, AlertGroupDefinition>;
    protected get defaultTestId(): string;
    protected get variantClassName(): string;
    protected get pluralize(): boolean;
    protected get oneContentItem(): string | undefined;
    protected get correctVariant(): AlertLevel | null;
    protected get groupOptions(): AlertGroupDefinition;
}
