import type RouterService from '@ember/routing/router-service';
import type RouteInfo from '@ember/routing/-private/route-info';
import type { IBreadCrumbController, BreadCrumb } from '@nsf-open/ember-ui-foundation/constants';
import Component from '@ember/component';
declare type FullBreadCrumb = Required<Exclude<BreadCrumb, string>>;
declare type EngineInfo = {
    fullName: string;
    instanceId: number;
    localFullName: string;
    mountPoint: string;
    name: string;
};
/**
 * The UiBreadCrumbs component will generate a breadcrumb trail of hyperlinks by
 * walking upwards from the current route to the topmost application route,
 * reading `breadCrumb` or `breadCrumbs` configuration off of each controller
 * as it goes.
 *
 * Controllers have the option of providing breadcrumb configuration in a couple
 * of ways. This is opt-in, so not every controller in the routing stack needs to
 * provide something.
 *
 * ## Examples
 *
 * A string `breadCrumb` will become the label of the breadcrumb. When required,
 * the link href will be generated from the controller's name.
 *
 * ```typescript
 * export default class UserController extends Controller {
 *   readonly breadCrumb = 'User';
 * }
 * ```
 *
 * The string `breadCrumb` can also be dynamic, based on the controller's model or
 * some other computed value.
 *
 * ```typescript
 * export default class UserController extends Controller {
 *   readonly model: UserRouteModel;
 *
 *   @readOnly('model.username')
 *   declare readonly breadCrumb: string;
 * }
 * ```
 *
 * If there are dynamic segments in the path, then a model will need to be supplied
 * to fill them. This works the same as supplying a model to a `<LinkTo>` element.
 *
 * ```typescript
 * export default class UserController extends Controller {
 *   readonly model: UserRouteModel;
 *
 *   @computed('model.{id,username}')
 *   get breadCrumb() {
 *     return {
 *       label: this.model.username,
 *       model: this.model.id,
 *     };
 *   };
 * }
 * ```
 *
 * It is also possible to provide a path to a completely different route, if required.
 * This is the value that you would provide to a `<LinkTo>` @route argument. The same
 * rules about supplying a model still apply.
 *
 * ```typescript
 * export default class UserController extends Controller {
 *   readonly model: UserRouteModel;
 *
 *   @computed('model.{id,username}')
 *   get breadCrumb() {
 *     return {
 *       label: this.model.username,
 *       model: this.model.id,
 *       path:  'some.other.route',
 *     };
 *   };
 * }
 * ```
 *
 * In some circumstances a single controller will need to describe multiple breadcrumb
 * segments. This is possible using the `breadCrumbs` (note, plural of `breadCrumb`)
 * property on a controller. An array can be provided now, with the same rules described
 * above still applying to each item.
 *
 * ```typescript
 * export default class UserController extends Controller {
 *   readonly model: UserRouteModel;
 *
 *   @computed('model.{id,username}')
 *   get breadCrumbs() {
 *     return [
 *       this.model.username,
 *       {
 *         label: 'Edit',
 *         model: this.model.id,
 *         path:  'users.user.edit',
 *       },
 *     ];
 *   };
 * ```
 *
 * In some even more rare circumstances a breadcrumb segment will need to link to a completely
 * custom URL that might not be part of the application at all. When an `href` value is provided
 * it will be provided as-is to the anchor. An optional `target` (default "_self") can be provided
 * as well.
 *
 * ```typescript
 * export default class UserController extends Controller {
 *   readonly model: UserRouteModel;
 *
 *   @computed('model.{id,username}')
 *   get breadCrumb() {
 *     return {
 *       label:  this.model.username,
 *       href:   'https://site.noexist/users/${this.model.id}',
 *       target: '_blank',
 *     };
 *   };
 * }
 * ```
 */
export default class UiBreadCrumbs extends Component {
    readonly router: RouterService;
    get crumbs(): {
        label?: string | undefined;
        path?: string | undefined;
        model?: unknown;
        linkable?: boolean | undefined;
        isCurrent?: boolean | undefined;
        rewind?: number | undefined;
        href?: string | undefined;
        target?: "_self" | "_top" | "_blank" | "_parent" | undefined;
    }[];
    buildSingleCrumb(crumb: BreadCrumb, routeInfo: RouteInfo): FullBreadCrumb;
    lookupController(name: string): IBreadCrumbController | null;
    lookupEngineInfoByRoute(name: string): EngineInfo | undefined;
    lookupEngineInstance(info: EngineInfo): any;
}
export {};
