import Component from '@ember/component';
import { ButtonVariants } from '../../constants';
declare enum ListNav {
    FIRST = 0,
    LAST = 1,
    PREV = 2,
    NEXT = 3
}
/**
 * Menus allow for several related actions/options to be grouped together and only made visible
 * when requested.
 *
 * ```js
 * <UiMenu @buttonText="Align" as |Menu|>
 *   <Menu.Item @text="Left"  @onClick={{action this.handleAlignSelection "left"}} />
 *   <Menu.Item @text="Right" @onClick={{action this.handleAlignSelection "right"}} />
 * </UiMenu>
 * ```
 *
 * The yielded `item` is actually a UiButton reference, so anything that you can do with that
 * you can also do here for your menu options.
 *
 * By default, the menu options flyout will be positioned below and right of the trigger button,
 * but will automatically pick another orientation if the trigger's placement requires it.
 *
 * Accessibility-wise the menu announces itself as such, and provides the expected navigation
 * using keys such as `arrow up`, `arrow down`, `escape`, `enter`, and `space bar`.
 *
 * @yields {UiMenuFlyout} item - A UiButton to create menu options with.
 */
export default class UiMenu extends Component {
    static readonly positionalParams: string[];
    /**
     * Boolean. Programmatically toggle the menu flyout.
     */
    visible: boolean;
    /**
     * The text content of the trigger button. This can also be set as the first positional parameter
     * of the component.
     */
    buttonText?: string;
    /**
     * The style variant of the control.
     */
    variant: ButtonVariants;
    /**
     * The font-awesome icon displayed on the menu toggle button.
     */
    icon: string;
    /**
     * The value of the element's `data-test-id` attribute, if required.
     */
    testId?: string;
    /**
     * Boolean. Disables interaction with the control.
     */
    disabled: boolean;
    private overlayElement;
    private overlayKeyListener;
    /**
     * @protected
     */
    triggerId: string;
    /**
     * @protected
     */
    overlayId: string;
    init(): void;
    didInsertElement(): void;
    willDestroyElement(): void;
    protected setVisibility(state: boolean): void;
    protected getOverlayElement(): HTMLElement | null;
    protected getMenuItems(): HTMLElement[];
    protected focusOnMenuItem(cmd: ListNav): void;
    /**
     * Moves document focus to the trigger button.
     *
     * @method focusOnTrigger
     * @protected
     */
    protected focusOnTrigger(): void;
    /**
     * @method handleOverlayShown
     * @protected
     */
    protected handleOverlayShown(): void;
    /**
     * @method handleOverlayHidden
     * @protected
     */
    protected handleOverlayHidden(): void;
    /**
     * The keyup event listener for the trigger button. Provides keyboard navigation
     * when focused on the open/close button.
     *
     * @method handleTriggerKeyEvents
     * @param {KeyboardEvent} event
     * @protected
     */
    protected handleTriggerKeyEvents(event: KeyboardEvent): void;
    /**
     * The keyup event listener for the menu overlay. Provides keyboard navigation
     * through the list of menu items.
     *
     * @method handleOverlayKeyEvents
     * @param {KeyboardEvent} event
     * @protected
     */
    protected handleOverlayKeyEvents(event: KeyboardEvent): void;
    /**
     * @method handleMenuItemClick
     * @protected
     */
    protected handleMenuItemClick(): void;
}
export {};
