import Component from '@ember/component';
/**
 * The `UiPager` component takes in an array of _things_ and turns it into
 * multiple smaller, sequential arrays that can be used for rendering, further
 * processing, or whatever else you might need.
 *
 * ```handlebars
 * <UiPager @records={{this.recordSet}} as |Pager|>
 *   {{!-- A <select> with options to control the max size of each page --}}
 *   <Pager.SizeOptions />
 *
 *   {{!-- Options to move around between pages --}}
 *   <Pager.Navbar />
 *
 *   {{#each Pager.pageRecords as |record|}}
 *     {{!-- Iterate through each record in the current slice --}}
 *   {{/each}}
 * </UiPager>
 * ```
 *
 * @yields (string) start          - The ones index of the first record of the current
 *                                  slice within the full recordset.
 * @yields (string) end            - The ones index of the last record of the current
 *                                   slice within the full recordset.
 * @yields (string) total          - The number of records within the full recordset.
 * @yields (number) pageSize       - The maximum number of records that will appear on
 *                                   a single page.
 * @yields (number) currentPage    - The ones index of the current page that is being
 *                                   provided by the component.
 * @yields (boolean) isPaged       - Whether paging is currently happening. This might
 *                                   be false, for example, if there are zero records in
 *                                   the full recordset, or if there is only one page
 *                                   worth or records.
 * @yields (unknown[]) pageRecords - The slice of the full recordset that constitutes the
 *                                   current page.
 * @yields (string) description    - The generated description of the pager state, by
 *                                   default it will read _"{start index} - {end index}
 *                                   of {total record count}"_.
 */
export default class UiPager extends Component {
    static readonly positionalParams: string[];
    /**
     * The source array to be paged.
     */
    records: Record<string, unknown>[];
    private _lastRecords?;
    /**
     * The maximum number of records that will be put into each page.
     */
    pageSize: string | number;
    private _lastPageSize?;
    /**
     * The 1's-based index of the current page.
     */
    currentPage: string | number;
    /**
     * An array of objects that get used to populate the page size selector.
     * By default, it provides 10, 50, and "All" options.
     */
    pageSizes: {
        value: string;
        label: string;
    }[];
    /**
     * If true, then the page size options presented will not exceed the length
     * of the record set. For example, if the `pageSizes` provided included 10,
     * 20, 30, 40, and 50, but the provided `records` array only contained 35 items,
     * then only the options for 10, 20, and 30 would be shown. An option with the
     * "Show All" value of -1 is exempt from this.
     */
    trimSizeOptions: boolean;
    /**
     * Whether the page selector can be interacted with.
     */
    disabled: boolean;
    /**
     * Callback that is uan when the current page changes.
     */
    onPageChange?: (newPageIndex: number) => void;
    /**
     * Callback that is run when the page size is changed.
     */
    onPageSizeChange?: (newPageSize: number) => void;
    /**
     * The method used to create the pager's state description. By default, the created
     * string reads _"{start index} - {end index} of {total record count}"_.
     */
    createDescription: (_page: number, start: number, end: number, total: number) => string;
    /**
     * The subset of pageSizes whose value is smaller than the length of the records array.
     */
    protected get availablePageSizes(): {
        value: string;
        label: string;
    }[];
    /**
     * The value of `pageSize` always cast as an integer.
     */
    protected get intPageSize(): number;
    /**
     * Keeps from having the first option in the list confusingly "selected" if the
     * number of records is changed to less than the page size value while `trimSizeOptions`
     * is enabled.
     *
     * Without a value, select elements visibly display their first option which might give
     * the wrong impression to the user. With `trimSizeOptions` it is possible to lose the
     * option that was selected and cause this visual glitch. The only way it occurs is when
     * the number of records dips below the selected page size, at which point the -1 "Show
     * All" option is applicable. Of course, if there is no "Show All" then you'll have an
     * empty select... developer's choice.
     */
    protected get selectedPageSize(): number;
    /**
     * The value of `currentPage` always cast as an integer.
     */
    protected get intCurrentPage(): number;
    /**
     * True if more than one page of content exists.
     */
    protected isPaged: boolean;
    /**
     * True if there are pages before the current index.
     */
    protected hasPrevPage: boolean;
    /**
     * True if there are pages after the current index.
     */
    protected get hasNextPage(): boolean;
    /**
     * The total number of records across all pages.
     */
    protected get recordCount(): number;
    /**
     * The total number of pages available.
     */
    protected get pageCount(): number;
    /**
     * The 1's-based index of the first item in `records` for the current page.
     */
    protected get startIndex(): number;
    /**
     * The 1's-based index of the last item in `records` for the current page.
     */
    protected get endIndex(): number;
    /**
     * The records contained in the current page, defined as the subset of `records`
     * between `startIndex` and `endIndex`.
     */
    protected get pageRecords(): Record<string, unknown>[] | null;
    /**
     * The current paging descriptor string.
     */
    protected get description(): string;
    init(): void;
    didUpdateAttrs(): void;
    /**
     * Navigate to a specific page by providing a 1's-based index.
     */
    protected goToPage(idx: number): void;
    /**
     * Navigate to the first page.
     */
    protected goToFirstPage(): void;
    /**
     * Navigate to the page prior to the current one, if available.
     */
    protected goToPrevPage(): void;
    /**
     * Navigate to the page after the current one, if available.
     */
    protected goToNextPage(): void;
    /**
     * Navigate to the last page.
     */
    protected goToLastPage(): void;
    /**
     * Change the maximum number of records on a page.
     */
    changePageSize(size: string | number): void;
}
