import type NativeArray from '@ember/array/-private/native-array';
import { AlertLevel, AlertLevelKeys } from '../constants';
export interface Message {
    id: string;
    message: string;
    escapeHTML: boolean;
    details: string | null;
    detailsOpen: boolean;
}
export interface MessageGroup {
    name: AlertLevel;
    messages: NativeArray<Message>;
    messagesText: Set<string>;
    clear: () => void;
}
export declare type StringOrStringArray = string | string[];
export declare type MessageMetadata = {
    hasSuccesses: boolean;
    hasErrors: boolean;
    hasWarnings: boolean;
    hasInformationals: boolean;
    warnings: string[];
    errors: string[];
    successes: string[];
    informationals: string[];
};
export declare function getCorrectedAlertLevel(name: string): AlertLevel | null;
export declare type MessageHash = Partial<Record<AlertLevelKeys, StringOrStringArray>>;
export declare enum MessageEvents {
    MESSAGE_ADDED = "received"
}
export declare const messageManager: (options?: {
    enableScrollTo?: boolean;
}) => import("@ember/object/computed").default<MessageManager, MessageManager>;
/**
 * The MessageManager acts as the interface for a UiAlertBlock.
 * Messages can be added, updated, and/or removed from an instance of this class, and the component
 * that it is attached to will update in response.
 *
 * @class MessageManager
 */
export default class MessageManager {
    enableDetails: boolean;
    /**
     * If true, the ui-alert-block that is rendering the contents of this instance will be scrolled
     * into view when the content changes.
     *
     * @property enableScrollTo
     * @type {boolean}
     */
    enableScrollTo: boolean;
    /**
     * An array of `group` objects. A group can be thought of as the "type" or "severity" of the messages
     * put into it (really though, these are arbitrary). By default, four groups are created for a new
     * instance of this class: `error`, `warning`, `information`, and `success`. The configuration for
     * each corresponds to the Bootstrap CSS "alert" banner: `alert-danger`, `alert-warning`,
     * `alert-secondary`, and `alert-success`.
     *
     * @property groups
     * @type MutableArray<MessageGroup>
     */
    readonly groups: NativeArray<MessageGroup>;
    constructor();
    /**
     * Retrieves the MessageGroup with the given name.
     *
     * @method getGroup
     *
     * @param {AlertLevel} name
     *
     * @return {MessageGroup | null}
     */
    getGroup(name: AlertLevel): MessageGroup | undefined;
    /**
     * Adds a new MessageGroup if one with the given name does not already exist.
     *
     * @method addGroup
     *
     * @param {AlertLevel} name
     *
     * @return {MessageGroup}
     */
    addGroup(name: AlertLevel): MessageGroup;
    /**
     * Does what it says. If a MessageGroup with the given name exists then it will be
     * returned, otherwise one will be created.
     *
     * @method getOrAddGroup
     *
     * @param {AlertLevel} name
     *
     * @return {MessageGroup}
     */
    getOrAddGroup(name: AlertLevel): MessageGroup;
    /**
     * Removes the MessageGroup with the given name.
     *
     * @method removeGroup
     *
     * @param {AlertLevel} name
     *
     * @return {boolean}
     */
    removeGroup(name: AlertLevel): boolean;
    /**
     * Check whether this MessageManager has the specified message string.
     * If a group is provided, only that provided group is checked;
     * else, all groups are checked.
     *
     * @method hasMessage
     *
     * @param {string} messageText     - The text of the message to check
     * @param {AlertLevel} groupName   - The group to look in, or `null` to look in all groups
     *
     * @return {boolean} Does this MessageManager have the specified message?
     *
     * @author  charvey@associates.nsf.gov
     * @updated 2019-12-19
     */
    hasMessage(messageText: string, groupName?: AlertLevel | null): boolean;
    /**
     * Create a Message and add it to the named MessageGroup
     *
     * @method addMessage
     *
     * @param {AlertLevel}  groupName          The name of the MessageGroup to add the new message to.
     * @param {string}      message            The text content of the message.
     *
     * @param {object}      [options]
     * @param {string}      [options.details=null]     Additional content to ride along with the message.
     * @param {boolean}     [options.clearPrior=false] Whether or not current messages in the group should be cleared before adding.
     * @param {boolean}     [options.escape=true]      Whether or not to HTML escape the contents of `message`.
     *
     * @return {string|boolean}
     */
    addMessage(groupName: AlertLevel, message: string, options?: {
        details?: string | null;
        clearPrior?: boolean;
        escape?: boolean;
    }): string | false;
    /**
     * Add many simple messages to a group.
     *
     * @method addMessagesMany
     *
     * @param {AlertLevel}  groupName - The name of the group to add to
     * @param {string}      messages  - An array of string messages
     * @param {object}      [options]
     * @param {boolean}     [options.clearPrior]
     * @param {boolean}     [options.escape=true] - Whether to HTML escape the contents of `message`.
     *
     * @return {string[]}
     *
     * @author  charvey@associates.nsf.gov
     * @updated 2019-12-19
     */
    addMessagesMany(groupName: AlertLevel, messages: string[], options?: {
        clearPrior?: boolean;
        escape?: boolean;
    }): string[];
    /**
     * A convenience method which calls either `addMessage` or `addMessagesMany` depending
     * on the provided argument being a single string or array of strings.
     *
     * @method addMessages
     *
     * @param {AlertLevel}  groupName - The name of the group to add to
     * @param {string}      messages  - A single string or array of string messages
     * @param {object}      [options]
     * @param {boolean}     [options.clearPrior]
     *
     * @return {string[]}
     */
    addMessages(groupName: AlertLevel, messages: StringOrStringArray, options?: {}): string[];
    /**
     * Given a hash of AlertLevel keys, and string or string array values each value will be mapped
     * to it's keyed message group.
     *
     * @method addMessagesByGroup
     *
     * @param {MessageHash} obj
     * @param {object}      [options]
     * @param {boolean}     [options.clearPrior]
     */
    addMessagesByGroup(obj: MessageHash, options?: {
        clearPrior?: boolean;
    }): void;
    /**
     * Added one or more messages to the success group.
     *
     * @method addSuccessMessages
     *
     * @param {StringOrStringArray} messages
     * @param {object}              [options]
     * @param {boolean}             [options.clearPrior]
     *
     * @return {string[]}
     */
    addSuccessMessages(messages: StringOrStringArray, options?: {
        clearPrior?: boolean;
    }): string[];
    /**
     * Added one or more messages to the warning group.
     *
     * @method addWarningMessages
     *
     * @param {StringOrStringArray} messages
     * @param {object}              [options]
     * @param {boolean}             [options.clearPrior]
     *
     * @return {string[]}
     */
    addWarningMessages(messages: StringOrStringArray, options?: {
        clearPrior?: boolean;
    }): string[];
    /**
     * Added one or more messages to the error group.
     */
    addErrorMessages(messages: StringOrStringArray, options?: {
        clearPrior?: boolean;
    }): string[];
    /**
     * Added one or more messages to the info group.
     */
    addInfoMessages(messages: StringOrStringArray, options?: {
        clearPrior?: boolean;
    }): string[];
    /**
     * Updates an existing Message with new content.
     */
    updateMessage(messageId: string, newMessage: string, newDetails?: string | null): boolean;
    /**
     * Removes the message with the given ID. IDs are provided for each message when calling
     * `.addMessage()` and `.addMessagesMany()`.
     */
    removeMessage(messageId: string): boolean;
    /**
     * A one-stop shop convenience method that will add, update, or remove a message based on the provided
     * arguments hash.
     *
     * messageId == null AND message != null THEN add new message (will also require `groupName` to be provided)
     * messageId != null AND message == null THEN remove message
     * messageId != null AND message != null THEN update message
     */
    message(options?: {
        messageId?: string;
        groupName?: AlertLevel;
        message?: string;
        details?: string;
    }): string | false | null;
    /**
     * Check whether a given group is empty (has no messages). If no group is given,
     * check all groups.
     */
    isEmpty(groupName?: AlertLevel): boolean;
    /**
     * Empty the MessageGroup with the given name(s), or all groups.
     */
    clear(groupName?: AlertLevel | AlertLevel[]): void;
    private _scheduleTriggerEvent;
    private _triggerEvent;
}
