import type ProgressManager from './ProgressManager';
import type { IProgressComponent } from './ProgressComponent';
/**
 * The ProgressItemDescriptor describes a basic object that can be provided to a
 * ProgressManager to create a ProgressItem instance.
 */
export declare type ProgressItemDescriptor<Data> = {
    /**
     * A simple, short descriptor of the progress item being defined. This will be
     * used to identify the item in things like chevron bars or ordered lists.
     */
    label: string;
    /**
     * A more verbose descriptor of the progress item being defined. This can be
     * defined in addition to `label` and will be used in things like panel
     * headings.
     * */
    title?: string;
    /**
     * A ProgressComponent class or string name where a component can be resolved.
     * Different workflow implementations will use this in different ways.
     */
    component?: string | IProgressComponent<Data>;
    /**
     * If true then the progress item will default to its `completed` state. For
     * example, this should be set to true for a step in the progression of a
     * workflow that displays a pre-populated form with correct values.
     */
    complete?: boolean;
    /**
     * If true then the progress item will never be either complete or incomplete.
     * This is particularly useful for informational or confirmation steps within
     * a workflow.
     */
    indeterminate?: boolean;
};
/**
 * A ProgressItem defines a single step within a workflow - a link in a chain when
 * defining linear paths, or a node in a web for more complex designs. ProgressItem
 * instances are created by a ProgressManager and carry stateful information about
 * their own internal status as well as their standing within the larger group of
 * all items being managed.
 *
 * @class ProgressItem
 */
export default class ProgressItem<Data> {
    constructor(descriptor: ProgressItemDescriptor<Data>, manager: ProgressManager<Data>);
    /**
     * A simple, short descriptor of the progress item being defined. This will be
     * used to identify the item in things like chevron bars or ordered lists.
     */
    label: string;
    /**
     * A more verbose descriptor of the progress item being defined. This can be
     * defined in addition to `label` and will be used in things like panel
     * headings.
     * */
    title?: string;
    /**
     * A ProgressComponent class or string name where a component can be resolved.
     * Different workflow implementations will use this in different ways.
     */
    component?: string | IProgressComponent<Data>;
    /**
     * If true, the flow is allowed to progress past this step item. This value
     * will be ignored if the step is `indeterminate`.
     * */
    complete: boolean;
    /**
     * If true then the progress item will never be either complete or incomplete.
     * This is particularly useful for informational or confirmation steps within
     * a workflow.
     */
    indeterminate: boolean;
    /**
     * True if the progress item is either complete or indeterminate.
     */
    readonly isComplete: boolean;
    /**
     * The ProgressManager that owns this ProgressItem.
     */
    readonly manager: ProgressManager<Data>;
    /**
     * The index of this ProgressItem within the queue of its owning ProgressManager.
     */
    get index(): number;
    /**
     * The ProgressItem indexed immediately before this one, or undefined if this
     * is the first item in the workflow.
     */
    get previousStep(): ProgressItem<Data> | undefined;
    /**
     * The ProgressItem indexed immediately after this one, or undefined if this
     * is the last item in the workflow.
     */
    get nextStep(): ProgressItem<Data> | undefined;
    /**
     * True if this is the current ProgressItem in the ProgressManager.
     */
    get isActive(): boolean;
    /**
     * True if the ProgressItem immediately before this one is the current item in
     * the ProgressManager.
     */
    get isPreviousActive(): boolean;
    /**
     * True if the ProgressItem immediately after this one is the current item in
     * the ProgressManager.
     */
    get isNextActive(): boolean;
    /**
     * True if the current item in the ProgressManager is indexed lower than this
     * one.
     */
    get isPastStepActive(): boolean;
    /**
     * True if the current item in the ProgressManager is indexed higher than this
     * one.
     */
    get isFutureStepActive(): boolean;
    /**
     * True if this step is allowed to be navigated to from somewhere else. Current
     * support is limited to a linear workflow - the ProgressItem immediately before
     * this one must have been marked as either complete or indeterminate.
     */
    get canBeLinkedTo(): boolean;
    /**
     * A string representation of the current item's completion state and location
     * within the workflow.
     */
    get status(): string;
    /**
     * Sets the `complete` flag on this step to true.
     */
    markComplete(): void;
    /**
     * Sets the `complete` flag on this item to false.
     */
    markIncomplete(): void;
    /**
     * Updates the `complete` value of this item. This method is particularly
     * useful as the argument of a UiForm's `onFormValidationChange` callback.
     */
    updateCompleteState(complete: boolean | {
        isFormValid: boolean;
    }): void;
}
