import type { ProgressItemDescriptor } from './ProgressItem';
import ProgressItem from './ProgressItem';
/**
 * Decorator to create a new ProgressManager instance.
 *
 * ```typescript
 * import type { ProgressManager } from '@nsf-open/ember-ui-foundation';
 * import { progressManager } from '@nsf-open/ember-ui-foundation';
 * // ...
 * @progressManager([{ label: 'Step A', component: 'steps/step-a' }])
 * public declare manager: ProgressManager;
 * ```
 */
export declare function progressManager<Data = Record<string, unknown>>(steps?: ProgressItemDescriptor<Data>[]): import("@ember/object/computed").default<ProgressManager<Data>, ProgressManager<Data>>;
/**
 * The ProgressManager acts as an interface for step-oriented workflows such as
 * the UiStepflow, and supporting components like UiProgressBar.
 *
 * ```typescript
 * import type { ProgressManager } from '@nsf-open/ember-ui-foundation';
 * import { progressManager } from '@nsf-open/ember-ui-foundation';
 * // ...
 * @progressManager([{ label: 'Step A', component: 'steps/step-a' }])
 * public declare manager: ProgressManager;
 * ```
 */
export default class ProgressManager<Data> {
    constructor(steps?: (ProgressItem<Data> | ProgressItemDescriptor<Data>)[]);
    /** The ordered steps of this StepFlow. */
    readonly steps: import("@ember/array/-private/native-array").default<ProgressItem<Data>>;
    /**
     * A single generic `data` object that is shared between all ProgressItems
     * and is available for whatever need is required of it. For example, a multipart
     * form that spans several workflow steps can all record their inputs here for
     * easy retrieval.
     */
    data?: Data;
    /**
     * The index of the current ProgressItem within the workflow.
     */
    currentStepIndex: number;
    /**
     * The total number of items within the workflow.
     */
    readonly totalStepCount: number;
    /**
     * If true, there is at least one ProgressItem before the current.
     */
    readonly hasPreviousStep: boolean;
    /**
     * If true, there is at least one more ProgressItem after the current.
     */
    readonly hasNextStep: boolean;
    /**
     * The ProgressItem at the current index.
     */
    get currentStep(): ProgressItem<Data>;
    /**
     * A reference to the ProgressItem prior to the current one.
     */
    get previousStep(): ProgressItem<Data> | undefined;
    /**
     * A reference to the ProgressItem after the current one.
     */
    get nextStep(): ProgressItem<Data> | undefined;
    /**
     * The number of workflow steps that have been completed so far.
     */
    get completedStepCount(): number;
    /**
     * If true, all ProgressItems have been marked complete.
     */
    get isComplete(): boolean;
    /**
     * Retrieves the ProgressItem at the specified index.
     */
    getStepAt(index: number): ProgressItem<Data> | undefined;
    /**
     * Retrieves the ProgressItem that is one step before the specified index.
     */
    getStepBefore(index: number): ProgressItem<Data> | undefined;
    /**
     * Retrieves the ProgressItem that is one step after the specified index.
     */
    getStepAfter(index: number): ProgressItem<Data> | undefined;
    /**
     * Action to navigate to the ProgressItem at a specific index.
     */
    goToStep(index: number): boolean;
    /**
     * Action to navigate forward one step.
     */
    goToNextStep(): boolean;
    /**
     * Action to navigate back one step.
     */
    goToPreviousStep(): boolean;
    /**
     * Add a new step to the end of the current list.
     */
    addStep(step: ProgressItem<Data> | ProgressItemDescriptor<Data>): void;
}
