/**
 * A Transform is a type of method that takes one type of value and turns it
 * into something else. In other words, it _transforms_ it. Yup. Several token
 * rules use transforms to change the raw string value from the query into
 * a more usable format, e.g. the string "true" being cast to the boolean true.
 */
declare type Transform = (a: string) => unknown;
/**
 * An Operation is a type of method that takes two arguments and compares them
 * in some way. The comparison and logical token rules define such operators to
 * perform the computation of what the token represents, e.g. the AND token has
 * an operator that will return the value of two arguments (a && b).
 */
declare type Operation = (a: unknown, b: unknown) => boolean;
/**
 * An Evaluation method is what we're trying to get out of all that is going on
 * here. Given a record, return true/false whether it meets the query  criteria.
 */
declare type Evaluation = (record: EvaluatedRecord) => boolean;
/**
 * An EvaluatedRecord is a generic type of object that is to be provided for
 * comparison against the query.
 */
declare type EvaluatedRecord = Record<string, unknown>;
/**
 * A SimpleToken is created for each matched TokenRule within a query string
 * as that query is tokenized. It's pretty basic info, but this is only the
 * first step in processing.
 */
export declare type SimpleToken = {
    name: TokenKeys;
    value: string;
    start: number;
    end: number;
};
/**
 * A TokenRule defines a single concept that is expressible within a query,
 * something like "AND". It describes the regex to recognize it, and some
 * bits on what it means or how it might be manipulated.
 */
export declare type TokenRule = {
    name: TokenKeys;
    category: TokenCategories;
    regex: string;
    transform?: Transform;
    operation?: Operation;
};
/**
 * Categories are more generalized buckets that each TokenRule falls into
 * so that rule descriptions don't get too unwieldy.
 */
export declare enum TokenCategories {
    Constant = "Constant",
    Variable = "Variable",
    Logical = "Logical",
    Comparison = "Comparison",
    Grouping = "Grouping"
}
/**
 * The TokenKeys enumeration contains the names of all the special concepts
 * that can be expressed within a query.
 */
export declare enum TokenKeys {
    StringLiteral = "StringLiteral",
    BooleanLiteral = "BooleanLiteral",
    NumberLiteral = "NumberLiteral",
    NullLiteral = "NullLiteral",
    UndefinedLiteral = "UndefinedLiteral",
    ColumnIdentifier = "ColumnIdentifier",
    LogicalAnd = "LogicalAnd",
    LogicalOr = "LogicalOr",
    EqualsComparison = "EqualsComparison",
    StartsWithComparison = "StartsWithComparison",
    EndsWithComparison = "EndsWithComparison",
    IncludesComparison = "IncludesComparison",
    NotEqualToComparison = "NotEqualToComparison",
    DoesNotStartWithComparison = "DoesNotStartWithComparison",
    DoesNotEndWithComparison = "DoesNotEndWithComparison",
    DoesNotIncludeComparison = "DoesNotIncludeComparison",
    IsLessThanComparison = "IsLessThanComparison",
    IsGreaterThanComparison = "IsGreaterThanComparison",
    LeftParen = "LeftParen",
    RightParen = "RightParen"
}
/**
 * The TokenRules array holds all TokenRule definitions, one for each TokenKey.
 * An important thing to note here that order is important since the regex snippets
 * are evaluated as they appear. Put a more general rule before a more specific one,
 * and you're liable to get weird results.
 */
export declare const TokenRules: TokenRule[];
/**
 * @class QueryParser
 */
export default class QueryParser {
    regex: RegExp;
    constructor();
    /**
     * @method evaluate
     */
    evaluate(query: string): Evaluation;
    /**
     * Breaks apart the query string into an array of SimpleToken objects.
     *
     * @method tokenize
     */
    tokenize(query: string): SimpleToken[];
    /**
     * @method buildEvalTree
     */
    protected recurseEvalNode(tokens: SimpleToken[], prevToken?: SimpleToken): Evaluation;
    /**
     * @method getTokenRule
     * @protected
     */
    protected getTokenRule(name: TokenKeys): TokenRule;
    /**
     * @method getTokenRulesByCategory
     * @protected
     */
    protected getTokenRulesByCategory(category: TokenCategories): TokenRule[];
    /**
     * @method evaluateToken
     * @protected
     */
    protected evaluateToken(token: SimpleToken, prevToken?: SimpleToken, validCategories?: Array<TokenCategories>, validKeys?: Array<TokenKeys>): TokenRule;
    /**
     * @method extractTokenSequence
     * @protected
     */
    protected extractTokenSequence(tokens: SimpleToken[], start: TokenKeys, end: TokenKeys): SimpleToken[] | undefined;
    /**
     * Returns a function that will pass the `left` and `right` arguments through the `operand`
     * and return the result.
     *
     * @method createEvaluationFunction
     * @protected
     */
    protected createEvaluationFunction(left: unknown, operand: Operation, right: unknown): Evaluation;
    /**
     * @method createMissingTokenMessage
     * @protected
     */
    protected createMissingTokenMessage(prevTokenOrIndex: SimpleToken | number, expectedTypes: Array<string>): string;
    /**
     * @method createInvalidTokenMessage
     * @protected
     */
    protected createInvalidTokenMessage(token: SimpleToken, prevTokenOrIndex: SimpleToken | number, expectedTypes: Array<string>): string;
}
export {};
