import { SortOrder } from '../constants';
declare type SortInstruction = {
    sortOn: string;
    direction?: SortOrder;
};
/**
 * A SortRule stores information about sorting criteria such as the
 * property of an object to sort against, and the current direction
 * of the sort.
 */
export default class SortRule {
    /**
     * An optional, human-friendly name for the sort rule. If not provided,
     * then the value of `sortOn` will be used. Use `displayName` to access.
     */
    name?: string;
    /**
     * If being used in the context of multiple sort rules, this will reflect
     * the index of the rule - the order in which it will be applied.
     */
    index?: number;
    /**
     * The name of the property whose value will be used for sorting.
     */
    sortOn?: string;
    /**
     * A string of comma separated values that will be used as 2nd, 3rd, to n-th
     * order child sorts in the order provided. This is useful in situations where
     * a large list of identical values might be sorted indeterminately - think a
     * list of names with lots of people sharing the same first name. Providing a
     * `subSortOn` of "lastName" would help to further group the sorted results.
     */
    subSortOn?: string;
    /**
     * A compliment to `subSortOn` that provides a way for sub-sort properties to
     * have their sort direction "pinned" to a specific value. By default, sub-sorts
     * follow the primary sort direction.
     */
    subSortDirection?: string;
    /**
     * A three-step sort direction cycle
     */
    threeStepCycle: boolean;
    /** @private */
    _direction: SortOrder;
    /** @private */
    _previousDirection: SortOrder;
    /**
     * The current direction of the sort - ascending, descending, or none.
     */
    get direction(): SortOrder;
    set direction(value: SortOrder);
    /**
     * The last sort direction. This can be used in conjunction with the `direction`
     * property to figure out what comes next if cycling through sort ordering options.
     */
    get previousDirection(): SortOrder;
    /**
     *
     */
    get nextDirection(): SortOrder;
    /**
     *
     */
    get enabled(): boolean;
    /**
     *
     */
    get displayName(): string;
    /**
     * Simplifies the current state of the rule, as well as any sub-sort rules,
     * to an array of `sortOn` and `direction` objects to be read like an ordered
     * set of instructions.
     */
    flatten(): Required<SortInstruction>[];
    updateToNextSortDirection(): void;
}
export {};
