import Service from '@ember/service';
import { Promise } from 'rsvp';
/**
 * The BackdropService provides a single mechanism for controlling the full window
 * overlay that appears behind modal windows, loading indicators, or whatever else
 * is needed. When the backdrop is "enabled", the document body cannot be scrolled
 * and its content is muted with a semi-transparent opacity layer.
 *
 * You probably won't ever need to interact with this directly. Good to know it
 * exists though.
 *
 * ```ts
 * import type { BackdropService } from '@nsf-open/ember-ui-foundation/services';
 * import { inject as service } from '@ember/service';
 * // ...
 * @service
 * declare readonly backdrop: BackdropService;
 * ```
 */
export default class BackdropService extends Service {
    /**
     * The id of the backdrop element.
     *
     * @protected
     */
    backdropId: string;
    /**
     * The calculated width of a scrollbar for the browser being used. When the backdrop
     * is enabled this is set as padding on the document body if needed to counteract
     * losing the vertical scrollbar. Not doing so would result in a very obvious and
     * unsightly horizontal jump of content.
     *
     * @protected
     */
    scrollbarWidth: number;
    /**
     * @protected
     */
    bodyOverflow: boolean;
    /**
     * @protected
     */
    originalBodyPad: string;
    /**
     * The HTMLElement playing the role of the backdrop, if it exists.
     */
    private get backdropElement();
    /**
     * Show the backdrop. The returned promise will resolve once the opacity
     * animation has completed, or immediately if animation is disabled.
     *
     * @param [animate] Whether or not the opacity of the backdrop should be animated.
     */
    open(animate?: boolean): Promise<void>;
    /**
     * Hide the backdrop. The returned promise will resolve once the backdrop HTMLElement
     * is removed and scrolling has been returned to the document body.
     *
     * @param [animate] Whether or not the opacity of the backdrop should be animated.
     */
    close(animate?: boolean): Promise<void>;
    private lockDocumentBody;
    private unlockDocumentBody;
    private createBackdropElement;
    private destroyBackdropElement;
    /**
     * @protected
     */
    init(): void;
    /**
     * Cleanup. Ensure that the backdrop is destroyed if this service is disposed of.
     *
     * @protected
     */
    destroy(): import("@ember/object/core").default;
}
declare module '@ember/service' {
    interface Registry {
        backdrop: BackdropService;
    }
}
